"""causy

Causal discovery made easy. Causy allows you to use and implement causal discovery algorithms with easy to use, extend and maintain pipelines. It is built based on pytorch which allows you to run the algorithms on CPUs as well as GPUs seamlessly.

Learn more at https://causy.dev.

"""

import importlib.metadata

__version__ = importlib.metadata.version("causy")
